/*
Copyright 2015-2016, Manuel Bichler
WWW: <http://dbai.tuwien.ac.at/proj/lpopt/>

This file is part of lpopt.

lpopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

lpopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lpopt.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <algorithm>
#include <memory>

#include "asp-domain.hpp"
#include "StringGenerator.hpp"
#include "safetyCharacterization.hpp"

using namespace std;

namespace domain {
  
  static Rule* optimizeInnerDomain(LiteralBody* body, const varset& headVars, StringGenerator& gen, const varset& globalVars); // fwd-declaration

  vector<Rule*>* NafLiteral::optimizeInnerDomains(StringGenerator& /*gen*/, const varset& /*globalVars*/)
  {
    return NULL;
  }

  void Program::optimizeInnerDomains(StringGenerator& gen)
  {
    unsigned int i = 0;
    while(i < statements->size())
    {
      for (vector<Statement *>::iterator it = statements->begin()+i; it != statements->end(); ++it)
      {
	vector<Rule*>* res = (*it)->optimizeInnerDomains(gen);
	if (res != NULL)
	{
 	  statements->insert(it+1, res->begin(), res->end());
	  delete res;
	  i += 2; // don't try to optimize current and new statement. Continue afterwards.
	  break;
	}
	++i;
      }
    }
  }

  vector<Rule*>* OptimizeStatement::optimizeInnerDomains(StringGenerator& /*gen*/)
  {
    return NULL;
  }
  
  vector<Rule*>* Body::optimizeInnerDomains(StringGenerator& gen)
  {
    vector<Rule*>* rules = new vector<Rule*>();
    for (vector<BodyElement *>::iterator it = elements->begin(); it != elements->end(); ++it)
    {
      vector<Rule*>* res = (*it)->optimizeInnerDomains(gen,globalVariables);
      if (res != NULL)
      {
	rules->insert(rules->end(), res->begin(), res->end());
	delete res;
      }
    }
    if (rules->empty())
    {
      delete rules;
      return NULL;
    }
    return rules;
  }
  
  vector<Rule*>* WeakConstraint::optimizeInnerDomains(StringGenerator& gen)
  {
    return body->optimizeInnerDomains(gen);
  }
  
  vector<Rule*>* Rule::optimizeInnerDomains(StringGenerator& gen)
  {
    vector<Rule*>* bodyRes = body->optimizeInnerDomains(gen);
    varset *globalVars = variables();
    vector<Rule*>* headRes = head->optimizeInnerDomains(gen, *globalVars);
    delete globalVars;
    if (bodyRes == NULL) return headRes;
    if (headRes == NULL) return bodyRes;
    bodyRes->insert(bodyRes->begin(), headRes->begin(), headRes->end());
    delete headRes;
    return bodyRes;
  }
  
  vector<Rule*>* Disjunction::optimizeInnerDomains(StringGenerator& /*gen*/, const varset& /*globalVars*/)
  {
    return NULL;
  }
  
  static Rule* optimizeInnerDomain(LiteralBody* body, const varset& headVars, StringGenerator& gen, const varset& globalVars) // to be used by AggregateElement and ChoiceElement
  {
    // split the body into two: those literals that contain at least one body-unsafe variable are kept in body, the others got into b, apart from unsafes in b-context.
    SafetyCharacterization* bodySafety = body->safety();
    LiteralBody* b = body->splitContains(bodySafety->unsafeVars());
    delete bodySafety;
    
    if (b->size() == 0) {
      delete b;
      return NULL;
    }
    if (b->size() == 1) {
      body->combine(b);
      return NULL;
    }
    
    // variables of new predicate are called predVars.
    varset* v = body->variables();
    v->insert(globalVars.begin(), globalVars.end());
    v->insert(headVars.begin(), headVars.end());
    // now, intersect v with those that are in body
    varset* bvar = b->variables();

    vector<const Variable*> predVars(bvar->size());
    vector<const Variable*>::iterator it = std::set_intersection(v->begin(), v->end(), bvar->begin(), bvar->end(), predVars.begin(), var_cmp());
    predVars.resize(it - predVars.begin());
    delete v;
    delete bvar;
    
    
    // create domain object from predVars
    ClassicalLiteral* newPred;
    if (predVars.empty())
    { // no term tuples possible
      newPred = new ClassicalLiteral(gen.get(), false, false);
    } else {
      vector<Term *> *terms = new vector<Term *>();
      for (it = predVars.begin(); it != predVars.end(); ++it)
	terms->push_back(new Variable(**it));
      newPred = new ClassicalLiteral(gen.get(), false, false, new TermTuples(new vector<TermTuple *>(1, new TermTuple(terms))));
    }
    
    // put newPred into the aggregate body
    body->insert(newPred);
    
    // construct new rule
    Rule *r = new Rule(new Disjunction(new vector<ClassicalLiteral *>(1, new ClassicalLiteral(*newPred))), b->toRegularBody());
    delete b;
    
    return r;
  }

  Rule* ChoiceElement::optimizeInnerDomain(StringGenerator& gen, const varset& globalVars)
  {
    return domain::optimizeInnerDomain(dependencyLiterals, ((BodyElement*)literal)->variablesFast(), gen, globalVars);
  }
  
  vector<Rule*>* Choice::optimizeInnerDomains(StringGenerator& gen, const varset& globalVars)
  {
    vector<Rule*>* rules = new vector<Rule*>();
    for (vector<ChoiceElement *>::iterator it = elements->begin(); it != elements->end(); ++it)
    {
      Rule* r = (*it)->optimizeInnerDomain(gen, globalVars);
      if (r != NULL)
      {
	rules->push_back(r);
      }
    }
    if (rules->empty())
    {
      delete rules;
      return NULL;
    }
    return rules;
  }
  
  Rule* AggregateElement::optimizeInnerDomain(StringGenerator& gen, const varset& globalVars)
  {
    unique_ptr<varset> headVars(termTuple == NULL ? new varset() : termTuple->variables());
    return domain::optimizeInnerDomain(dependencyLiterals, *headVars, gen, globalVars);
  }
  
  vector<Rule*>* Aggregate::optimizeInnerDomains(StringGenerator& gen, const varset& globalVars)
  {
    vector<Rule*>* rules = new vector<Rule*>();
    for (vector<AggregateElement *>::iterator it = elements->begin(); it != elements->end(); ++it)
    {
      Rule* r = (*it)->optimizeInnerDomain(gen, globalVars);
      if (r != NULL)
      {
	rules->push_back(r);
      }
    }
    if (rules->empty())
    {
      delete rules;
      return NULL;
    }
    return rules;
  }
  

  
} // domain

 
