/*
Copyright 2015-2016, Manuel Bichler
WWW: <http://dbai.tuwien.ac.at/proj/lpopt/>

This file is part of lpopt.

lpopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

lpopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lpopt.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "asp-domain.hpp"
//TODO: optimization: replace delete-after-clone to "set null, clone, set object"
//TODO: optimization: clones out of order
// In this file, the splitPooling* methods are defined that split elements into multiple ones.

using namespace std;
using namespace domain;

vector<TermTuple*>* TermTuple::splitPooling()
{ // following algorithm works for splitPooling methods of all classes that hold a vector of subobjects and that have copying semantics in splitPooling
  vector<TermTuple*>* additionalTuples = new vector<TermTuple*>();
  for(vector<Term *>::iterator tupleit = terms->begin(); tupleit!=terms->end(); ++tupleit) // iterate over the term tuple
  {
    vector<Term*>* res = (*tupleit)->splitPooling(); // split the current term
    for(vector<Term *>::iterator termsit = res->begin(); termsit!=res->end(); ++termsit) // iterate over the alternative terms of the current term
    {
      // add a copy of this to additionalTuples with the current term replaced by the current alternative:
      Term* currentTermBackup = *tupleit;
      *tupleit = *termsit;
      TermTuple* copy = this->clone();
      delete *termsit;
      *tupleit = currentTermBackup;
      additionalTuples->insert(additionalTuples->begin(), copy); // TODO here it gets out of order: we do not want to add at the beginning, nor at the end --- it depends on the number of terms so far (or in between)
    }
    delete res;
  }
  
  vector<TermTuple*>* returnVector = new vector<TermTuple*>();
  for(vector<TermTuple *>::iterator addit = additionalTuples->begin(); addit!=additionalTuples->end(); ++addit) // iterate over the additional TermTuples
  {
    returnVector->push_back(*addit);
    vector<TermTuple*>* nonpoolingTuples = (*addit)->splitPooling(); // recursive pooling splitting
    returnVector->insert(returnVector->end(), nonpoolingTuples->begin(), nonpoolingTuples->end()); // adds all results to returnVector
    delete nonpoolingTuples;
  }
  delete additionalTuples;
  return returnVector;
}

vector<TermTuples*>* TermTuples::splitPooling()
{
  vector<TermTuples*>* returnVector = new vector<TermTuples*>();
  if (tuples->size() == 1)
  { // no pooling in this hierarchy, just call splitPooling on our single TermTuple and create a new TermTuples object with each result
    vector<TermTuple*>* subSplit = (*tuples)[0]->splitPooling();
    for (vector<TermTuple*>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
      returnVector->push_back(new TermTuples(new vector<TermTuple*>(1, *subSplitIt)));
    delete subSplit;
    return returnVector;
  }

  // we have pooling in this hierarchy. use a recursive approach: split last tuple, remove it from this and call this->splitPooling recursively.
  vector<TermTuple*>* subSplit = tuples->back()->splitPooling();
  returnVector->push_back(new TermTuples(new vector<TermTuple*>(1,tuples->back()))); // add last term tuple as a new TermTuples object
  for (vector<TermTuple*>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
    returnVector->push_back(new TermTuples(new vector<TermTuple*>(1, *subSplitIt))); // add all alternatives of last term tuple as a new TermTuples object
  delete subSplit;
  tuples->pop_back(); // remove last element, since it and its alternatives are already in returnVector
  vector<TermTuples*>* recursiveAdditions = this->splitPooling();
  returnVector->insert(returnVector->begin(), recursiveAdditions->begin(), recursiveAdditions->end());
  delete recursiveAdditions;
  return returnVector;
}

vector<Term*>* UnaryTerm::splitPooling()
{
  vector<Term*>* returnVector = new vector<Term*>();
  vector<Term*>* subSplit = base->splitPooling();
  for (vector<Term*>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
    returnVector->push_back(new UnaryTerm(*subSplitIt, type)); // add all alternatives as a new UnaryTerm object
  delete subSplit;
  return returnVector;
}

vector<Term*>* BracketedTerm::splitPooling()
{
  vector<Term*>* returnVector = new vector<Term*>();
  vector<Term*>* subSplit = base->splitPooling();
  for (vector<Term*>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
    returnVector->push_back(new BracketedTerm(*subSplitIt)); // add all alternatives as a new BracketedTerm object
  delete subSplit;
  return returnVector;
}

vector<Term*>* BinaryTerm::splitPooling()
{
  vector<BinaryTerm*>* additionalBinaryTerms = new vector<BinaryTerm*>();
  
  // first, split left term:
  vector<Term*>* subSplit = left->splitPooling();
  for(vector<Term *>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
  {
    // add a copy of this to additionalBinaryTerms with the left term replaced by the current alternative *subSplitIt:
    Term* leftBackup = left;
    left = *subSplitIt;
    BinaryTerm* copy = this->clone();
    delete *subSplitIt;
    left = leftBackup;
    additionalBinaryTerms->push_back(copy);
  }
  delete subSplit;
  // now, split right term:
  subSplit = right->splitPooling();
  for(vector<Term *>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
  {
    // add a copy of this to additionalBinaryTerms with the right term replaced by the current alternative *subSplitIt:
    Term* rightBackup = right;
    right = *subSplitIt;
    BinaryTerm* copy = this->clone();
    delete *subSplitIt;
    right = rightBackup;
    additionalBinaryTerms->push_back(copy);
  }
  delete subSplit;
  
  vector<Term*>* returnVector = new vector<Term*>();
  for(vector<BinaryTerm *>::iterator addit = additionalBinaryTerms->begin(); addit!=additionalBinaryTerms->end(); ++addit) // iterate over the additional terms
  {
    returnVector->push_back(*addit);
    vector<Term*>* nonpoolingTuples = (*addit)->splitPooling(); // recursive pooling splitting
    returnVector->insert(returnVector->end(), nonpoolingTuples->begin(), nonpoolingTuples->end()); // adds all results to returnVector
    delete nonpoolingTuples;
  }
  delete additionalBinaryTerms;
  return returnVector;
}

vector<Term*>* PoolingTerm::splitPooling()
{
  if (terms->size() == 1) // this is an already split pooling term
  {
    vector<Term*>* subTerms = (*terms)[0]->splitPooling();
    if((*terms)[0]->bracketingRelevant())
    {
      (*terms)[0] = new BracketedTerm((*terms)[0]);
    }
    for(vector<Term*>::iterator it = subTerms->begin(); it!=subTerms->end(); ++it)
    {
      if((*it)->bracketingRelevant())
      {
	(*it) = new BracketedTerm(*it);
      }
    }
    return subTerms;
  }
  // we have pooling in this hierarchy. use a recursive approach: split last term, remove it from this and call this->splitPooling recursively.
  vector<Term*>* returnVector = new vector<Term*>();
  vector<Term*>* subSplit = terms->back()->splitPooling();
  if(terms->back()->bracketingRelevant())
  {
    Term* brTerm = new BracketedTerm(terms->back());
    terms->pop_back();
    terms->push_back(brTerm);
  }
  for(vector<Term*>::iterator it = subSplit->begin(); it!=subSplit->end(); ++it)
  {
    if((*it)->bracketingRelevant())
    {
      (*it) = new BracketedTerm(*it);
    }
  }

  returnVector->push_back(terms->back()); // add last term
  returnVector->insert(returnVector->end(), subSplit->begin(), subSplit->end()); // add last term's alternatives
  delete subSplit;
  terms->pop_back(); // remove last term, since it and its alternatives are already in returnVector
  vector<Term*>* recursiveAdditions = this->splitPooling();
  returnVector->insert(returnVector->begin(), recursiveAdditions->begin(), recursiveAdditions->end());
  delete recursiveAdditions;
  return returnVector;
}

vector<Term*>* Constant::splitPooling()
{
  return new vector<Term*>();
}

vector<Term*>* Variable::splitPooling()
{
  return new vector<Term*>();
}

vector<Term*>* AnonymousVariable::splitPooling()
{
  return new vector<Term*>();
}

vector<Term*>* Function::splitPooling()
{
  vector<Term*>* returnVector = new vector<Term*>();
  vector<TermTuples*>* subSplit = tuples->splitPooling();
  for (vector<TermTuples*>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
  {
    TermTuples* backup = tuples;
    tuples = *subSplitIt;
    Function* copy = this->clone();
    delete *subSplitIt;
    tuples = backup;
    returnVector->push_back(copy);
  }
  delete subSplit;
  return returnVector;
}

template <typename T>
vector<T *> *ClassicalLiteral::splitPooling()
{
  static_assert(
    std::is_base_of<T, ClassicalLiteral>::value, 
    "T must be a supertype of ClassicalLiteral"
  );
  vector<T*>* returnVector = new vector<T*>();
  if (tuples==NULL) return returnVector;
  
  vector<TermTuples*>* subSplit = tuples->splitPooling();
  for (vector<TermTuples*>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
  {
    TermTuples* backup = tuples;
    tuples = *subSplitIt;
    ClassicalLiteral* copy = this->clone();
    delete *subSplitIt;
    tuples = backup;
    returnVector->push_back(copy);
  }
  delete subSplit;
  updatePublicVariables();
  return returnVector;
}

vector<NafLiteral*>* ClassicalLiteral::splitPoolingNaf()
{
  return splitPooling<NafLiteral>();
}

vector<BodyElement*>* ClassicalLiteral::splitPooling()
{
  return splitPooling<BodyElement>();
}

void Program::splitPooling()
{
  vector<Statement*>* splitStatements = new vector<Statement*>();
  for (vector<Statement*>::iterator statementsIt = statements->begin(); statementsIt!=statements->end(); ++statementsIt)
  {
    vector<Statement*>* sSplit = (*statementsIt)->splitPooling();
    splitStatements->push_back(*statementsIt);
    splitStatements->insert(splitStatements->end(), sSplit->begin(), sSplit->end());
    delete sSplit;
  }
  delete this->statements;
  this->statements = splitStatements;
}

vector<WeightAtLevel*>* WeightAtLevel::splitPooling()
{
  vector<WeightAtLevel*>* additionalWeightAtLevel = new vector<WeightAtLevel*>();
  
  // first, split weight term:
  vector<Term*>* subSplit = weight->splitPooling();
  for(vector<Term *>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
  {
    // add a copy of this to additionalWeightAtLevel with the weight term replaced by the current alternative *subSplitIt:
    Term* backup = weight;
    weight = *subSplitIt;
    WeightAtLevel* copy = this->clone();
    delete *subSplitIt;
    weight = backup;
    additionalWeightAtLevel->push_back(copy);
  }
  delete subSplit;
  
  // now, split level term:
  if (level != NULL)
  {
    subSplit = level->splitPooling();
    for(vector<Term *>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
    {
      // add a copy of this to additionalWeightAtLevel with the level term replaced by the current alternative *subSplitIt:
      Term* backup = level;
      level = *subSplitIt;
      WeightAtLevel* copy = this->clone();
      delete *subSplitIt;
      level = backup;
      additionalWeightAtLevel->push_back(copy);
    }
    delete subSplit;
  }
  
  // now, split tuple:
  if (tuple != NULL)
  {
    vector<TermTuple*>* subSplit2 = tuple->splitPooling();
    for(vector<TermTuple *>::iterator subSplitIt = subSplit2->begin(); subSplitIt!=subSplit2->end(); ++subSplitIt)
    {
      // add a copy of this to additionalWeightAtLevel with the tuple replaced by the current alternative *subSplitIt:
      TermTuple* backup = tuple;
      tuple = *subSplitIt;
      WeightAtLevel* copy = this->clone();
      delete *subSplitIt;
      tuple = backup;
      additionalWeightAtLevel->push_back(copy);
    }
    delete subSplit2;
  }
  
  vector<WeightAtLevel*>* returnVector = new vector<WeightAtLevel*>();
  for(vector<WeightAtLevel *>::iterator addit = additionalWeightAtLevel->begin(); addit!=additionalWeightAtLevel->end(); ++addit) // iterate over the additional w@ls
  {
    returnVector->push_back(*addit);
    vector<WeightAtLevel*>* nonpooling = (*addit)->splitPooling(); // recursive pooling splitting
    returnVector->insert(returnVector->end(), nonpooling->begin(), nonpooling->end()); // adds all results to returnVector
    delete nonpooling;
  }
  delete additionalWeightAtLevel;
  return returnVector;
}

vector<OptimizeElement*>* OptimizeElement::splitPooling()
{
  vector<OptimizeElement*>* additionalOptimizeElements = new vector<OptimizeElement*>();
  
  // first, split weight@level:
  vector<WeightAtLevel*>* subSplit = weightAtLevel->splitPooling();
  for(vector<WeightAtLevel *>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
  {
    // add a copy of this to additionalOptimizeElements with the w@l replaced by the current alternative *subSplitIt:
    WeightAtLevel* backup = weightAtLevel;
    weightAtLevel = *subSplitIt;
    OptimizeElement* copy = this->clone();
    delete *subSplitIt;
    weightAtLevel = backup;
    additionalOptimizeElements->push_back(copy);
  }
  delete subSplit;
  
  // now, split literals:
  vector<LiteralBody *> *subSplit2 = literals->splitPooling();
  for(vector<LiteralBody *>::iterator subSplitIt = subSplit2->begin(); subSplitIt!=subSplit2->end(); ++subSplitIt)
  {
    // add a copy of this to additionalOptimizeElements with literals replaced by the current alternative *subSplitIt:
    LiteralBody* backup = this->literals;
    this->literals = NULL;
    OptimizeElement* copy = this->clone();
    this->literals = backup;
    copy->literals = *subSplitIt;
    additionalOptimizeElements->push_back(copy);
  }
  delete subSplit2;
  
  vector<OptimizeElement*>* returnVector = new vector<OptimizeElement*>();
  for(vector<OptimizeElement *>::iterator addit = additionalOptimizeElements->begin(); addit!=additionalOptimizeElements->end(); ++addit) // iterate over the additional optimize elements
  {
    returnVector->push_back(*addit);
    vector<OptimizeElement*>* nonpoolingTuples = (*addit)->splitPooling(); // recursive pooling splitting
    returnVector->insert(returnVector->end(), nonpoolingTuples->begin(), nonpoolingTuples->end()); // adds all results to returnVector
    delete nonpoolingTuples;
  }
  delete additionalOptimizeElements;
  return returnVector;
}

vector<Statement*>* OptimizeStatement::splitPooling()
{
  if (elements->size() == 0) return new vector<Statement*>();
  
  vector<Statement*>* additionalStatements = new vector<Statement*>();
  if(elements->size() > 1)
  {
    // first, bring each optimize element into its own optimize statement:
    for(vector<OptimizeElement *>::iterator it = elements->begin() + 1 /*don't copy first*/; it!=elements->end(); ++it)
    {
      vector<OptimizeElement *>* backup = this->elements;
      this->elements = new vector<OptimizeElement *>(1, *it);
      OptimizeStatement *copy = this->clone();
      delete this->elements;
      this->elements = backup;
      additionalStatements->push_back(copy);
      vector<Statement*>* copyAlts = copy->splitPooling(); // recursive call
      additionalStatements->insert(additionalStatements->end(), copyAlts->begin(), copyAlts->end());
      delete copyAlts;
    }
    elements->resize(1); // remove all elements but the first since they have been outsourced into their own statements
  }
  
  // now, split pooling of this:
  vector<Statement*>* splitStatements = new vector<Statement*>();
  vector<OptimizeElement *>* subSplit = (*elements)[0]->splitPooling();
  for(vector<OptimizeElement *>::iterator it = subSplit->begin(); it!=subSplit->end(); ++it)
  {
    vector<OptimizeElement *>* backup = this->elements;
    this->elements = NULL;
    OptimizeStatement *copy = this->clone();
    this->elements = backup;
    copy->elements = new vector<OptimizeElement *>(1, *it);
    splitStatements->push_back(copy);
  }
  additionalStatements->insert(additionalStatements->end(), splitStatements->begin(), splitStatements->end());
  delete splitStatements;
  delete subSplit;
  return additionalStatements;
}

vector<Body*>* Body::splitPooling()
{
  vector<Body*>* additionalBodies = new vector<Body*>();
  for(vector<BodyElement *>::iterator elementsit = elements->begin(); elementsit!=elements->end(); ++elementsit) // iterate over the body elements
  {
    vector<BodyElement*>* res = (*elementsit)->splitPooling(); // split the current element
    for(vector<BodyElement *>::iterator altit = res->begin(); altit!=res->end(); ++altit) // iterate over the alternative elements of the current element
    {
      // add a copy of this to additionalBodies with the current element replaced by the current alternative:
      BodyElement* backup = *elementsit;
      *elementsit = *altit;
      Body* copy = this->clone();
      delete *altit;
      *elementsit = backup;
      additionalBodies->push_back(copy);
    }
    delete res;
  }
  
  vector<Body*>* returnVector = new vector<Body*>();
  for(vector<Body *>::iterator addit = additionalBodies->begin(); addit!=additionalBodies->end(); ++addit) // iterate over the additional bodies
  {
    returnVector->push_back(*addit);
    vector<Body*>* nonpoolingBodies = (*addit)->splitPooling(); // recursive pooling splitting
    returnVector->insert(returnVector->end(), nonpoolingBodies->begin(), nonpoolingBodies->end()); // adds all results to returnVector
    delete nonpoolingBodies;
  }
  delete additionalBodies;
  calcVariables();
  return returnVector;
}

vector<Statement*>* WeakConstraint::splitPooling()
{
  vector<WeakConstraint*>* additionalConstraints = new vector<WeakConstraint*>();
  
  // first, split body:
  vector<Body*>* subSplit = body->splitPooling();
  for(vector<Body *>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
  {
    // add a copy of this to additionalConstraints with the body replaced by the current alternative *subSplitIt:
    Body* backup = body;
    body = *subSplitIt;
    WeakConstraint* copy = this->clone();
    delete *subSplitIt;
    body = backup;
    additionalConstraints->push_back(copy);
  }
  delete subSplit;
  
  // now, split weightAtLevel:
  vector<WeightAtLevel*>* subSplit2 = weightAtLevel->splitPooling();
  for(vector<WeightAtLevel *>::iterator subSplitIt = subSplit2->begin(); subSplitIt!=subSplit2->end(); ++subSplitIt)
  {
    // add a copy of this to additionalConstraints with the weightAtLevel replaced by the current alternative *subSplitIt:
    WeightAtLevel* backup = weightAtLevel;
    weightAtLevel = *subSplitIt;
    WeakConstraint* copy = this->clone();
    delete *subSplitIt;
    weightAtLevel = backup;
    additionalConstraints->push_back(copy);
  }
  delete subSplit2;
  
  vector<Statement*>* returnVector = new vector<Statement*>();
  for(vector<WeakConstraint *>::iterator addit = additionalConstraints->begin(); addit!=additionalConstraints->end(); ++addit) // iterate over the additional constraints
  {
    returnVector->push_back(*addit);
    vector<Statement*>* nonpooling = (*addit)->splitPooling(); // recursive pooling splitting
    returnVector->insert(returnVector->end(), nonpooling->begin(), nonpooling->end()); // adds all results to returnVector
    delete nonpooling;
  }
  delete additionalConstraints;
  return returnVector;
}

vector<Statement*>* Rule::splitPooling()
{
  vector<Rule*>* additionalRules = new vector<Rule*>();
  
  // first, split body:
  vector<Body*>* subSplit = body->splitPooling();
  for(vector<Body *>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
  {
    // add a copy of this to additionalRules with the body replaced by the current alternative *subSplitIt:
    Body* backup = body;
    body = *subSplitIt;
    Rule* copy = this->clone();
    delete *subSplitIt;
    body = backup;
    additionalRules->push_back(copy);
  }
  delete subSplit;
  
  // now, split head:
  vector<RuleHead*>* subSplit2 = head->splitPooling();
  for(vector<RuleHead *>::iterator subSplitIt = subSplit2->begin(); subSplitIt!=subSplit2->end(); ++subSplitIt)
  {
    // add a copy of this to additionalRules with the head replaced by the current alternative *subSplitIt:
    RuleHead* backup = head;
    head = *subSplitIt;
    Rule* copy = this->clone();
    delete *subSplitIt;
    head = backup;
    additionalRules->push_back(copy);
  }
  delete subSplit2;
  
  vector<Statement*>* returnVector = new vector<Statement*>();
  for(vector<Rule *>::iterator addit = additionalRules->begin(); addit!=additionalRules->end(); ++addit) // iterate over the additional rules
  {
    returnVector->push_back(*addit);
    vector<Statement*>* nonpooling = (*addit)->splitPooling(); // recursive pooling splitting
    returnVector->insert(returnVector->end(), nonpooling->begin(), nonpooling->end()); // adds all results to returnVector
    delete nonpooling;
  }
  delete additionalRules;
  return returnVector;
}

vector<RuleHead*>* Disjunction::splitPooling()
{
  vector<Disjunction*>* additionals = new vector<Disjunction*>();
  for(vector<ClassicalLiteral *>::iterator literalsit = literals->begin(); literalsit!=literals->end(); ++literalsit) // iterate over the literals
  {
    vector<ClassicalLiteral*>* res = (*literalsit)->splitPooling<ClassicalLiteral>(); // split the current literal
    for(vector<ClassicalLiteral *>::iterator altit = res->begin(); altit!=res->end(); ++altit) // iterate over the alternative literals of the current literal
    {
      // add a copy of this to additionals with the current literal replaced by the current alternative:
      ClassicalLiteral* backup = *literalsit;
      *literalsit = *altit;
      Disjunction* copy = this->clone();
      delete *altit;
      *literalsit = backup;
      additionals->push_back(copy);
    }
    delete res;
  }
  
  vector<RuleHead*>* returnVector = new vector<RuleHead*>();
  for(vector<Disjunction *>::iterator addit = additionals->begin(); addit!=additionals->end(); ++addit) // iterate over the additional disjunctions
  {
    returnVector->push_back(*addit);
    vector<RuleHead*>* nonpoolingBodies = (*addit)->splitPooling(); // recursive pooling splitting
    returnVector->insert(returnVector->end(), nonpoolingBodies->begin(), nonpoolingBodies->end()); // adds all results to returnVector
    delete nonpoolingBodies;
  }
  delete additionals;
  return returnVector;
}

vector<Guard*>* Guard::splitPooling()
{
  vector<Guard*>* returnVector = new vector<Guard*>();
  vector<Term*>* subSplit = term->splitPooling();
  for (vector<Term*>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
  {
    // add a copy of this to returnVector with the current term replaced by the current alternative:
    Term* backup = this->term;
    this->term = NULL;
    Guard* copy = this->clone();
    this->term = backup;
    copy->term = (*subSplitIt);
    returnVector->push_back(copy);
  }
  delete subSplit;
  return returnVector;
}

vector<ChoiceElement*>* ChoiceElement::splitPooling()
{
  vector<ChoiceElement*>* additionalChoiceElements = new vector<ChoiceElement*>();
  
  // first, split literal:
  vector<ClassicalLiteral*>* subSplit = literal->splitPooling<ClassicalLiteral>();
  for(vector<ClassicalLiteral *>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
  {
    // add a copy of this to additionalChoiceElements with the literal replaced by the current alternative *subSplitIt:
    ClassicalLiteral* backup = this->literal;
    this->literal = NULL;
    ChoiceElement* copy = this->clone();
    this->literal = backup;
    copy->literal = *subSplitIt;
    additionalChoiceElements->push_back(copy);
  }
  delete subSplit;
  
  // now, split dependency literals:
  vector<LiteralBody *> *subSplit2 = dependencyLiterals->splitPooling();
  for(vector<LiteralBody *>::iterator subSplitIt = subSplit2->begin(); subSplitIt!=subSplit2->end(); ++subSplitIt)
  {
    // add a copy of this to additionalChoiceElements with the dependencyLiterals replaced by the current alternative *subSplitIt:
    LiteralBody* backup = this->dependencyLiterals;
    this->dependencyLiterals = NULL;
    ChoiceElement* copy = this->clone();
    this->dependencyLiterals = backup;
    copy->dependencyLiterals = *subSplitIt;
    additionalChoiceElements->push_back(copy);
  }
  delete subSplit2;
  
  vector<ChoiceElement*>* returnVector = new vector<ChoiceElement*>();
  for(vector<ChoiceElement *>::iterator addit = additionalChoiceElements->begin(); addit!=additionalChoiceElements->end(); ++addit) // iterate over the additional choice elements
  {
    returnVector->push_back(*addit);
    vector<ChoiceElement*>* nonpoolingChoiceElements = (*addit)->splitPooling(); // recursive pooling splitting
    returnVector->insert(returnVector->end(), nonpoolingChoiceElements->begin(), nonpoolingChoiceElements->end()); // adds all results to returnVector
    delete nonpoolingChoiceElements;
  }
  delete additionalChoiceElements;
  return returnVector;
}

vector<RuleHead*>* Choice::splitPooling()
{
  // first, split the elements *inside* this choice rule:
  vector<ChoiceElement*>* splitElements = new vector<ChoiceElement*>();
  for (vector<ChoiceElement*>::iterator elementsIt = elements->begin(); elementsIt!=elements->end(); ++elementsIt)
  {
    vector<ChoiceElement*>* sSplit = (*elementsIt)->splitPooling();
    splitElements->push_back(*elementsIt);
    splitElements->insert(splitElements->end(), sSplit->begin(), sSplit->end());
    delete sSplit;
  }
  delete this->elements;
  this->elements = splitElements;
  
  
  //now, handle guard poolings as usual, splitting into multiple Choices:
  vector<Choice*>* additionalChoices = new vector<Choice*>();
  
  // first, split left guard:
  if (guard_l != NULL)
  {
    vector<Guard*>* subSplit = guard_l->splitPooling();
    for(vector<Guard *>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
    {
      // add a copy of this to additionalChoices with the left guard replaced by the current alternative *subSplitIt:
      Guard* backup = this->guard_l;
      this->guard_l = NULL;
      Choice* copy = this->clone();
      this->guard_l = backup;
      copy->guard_l = *subSplitIt;
      additionalChoices->push_back(copy);
    }
    delete subSplit;
  }
  // now, split right guard:
  if (guard_r != NULL)
  {
    vector<Guard*>* subSplit = guard_r->splitPooling();
    for(vector<Guard *>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
    {
      // add a copy of this to additionalChoices with the right guard replaced by the current alternative *subSplitIt:
      Guard* backup = this->guard_r;
      this->guard_r = NULL;
      Choice* copy = this->clone();
      this->guard_r = backup;
      copy->guard_r = *subSplitIt;
      additionalChoices->push_back(copy);
    }
    delete subSplit;
  }
  
  vector<RuleHead*>* returnVector = new vector<RuleHead*>();
  for(vector<Choice *>::iterator addit = additionalChoices->begin(); addit!=additionalChoices->end(); ++addit) // iterate over the additional choice rules
  {
    returnVector->push_back(*addit);
    vector<RuleHead*>* nonpoolingTuples = (*addit)->splitPooling(); // recursive pooling splitting
    returnVector->insert(returnVector->end(), nonpoolingTuples->begin(), nonpoolingTuples->end()); // adds all results to returnVector
    delete nonpoolingTuples;
  }
  delete additionalChoices;
  return returnVector;
}

template <typename T>
vector<T *> *BuiltinAtom::splitPooling() // T must be supertype of BuiltinAtom
{
  static_assert(
    std::is_base_of<T, BuiltinAtom>::value, 
    "T must be a supertype of BuiltinAtom"
  );
  
  vector<BuiltinAtom*>* additionalAtoms = new vector<BuiltinAtom*>();
  
  // first, split left term:
  vector<Term*>* subSplit = t1->splitPooling();
  for(vector<Term *>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
  {
    // add a copy of this to additionalAtoms with the left term replaced by the current alternative *subSplitIt:
    Term* backup = this->t1;
    this->t1 = NULL;
    BuiltinAtom* copy = this->clone();
    this->t1 = backup;
    copy->t1 = *subSplitIt;
    additionalAtoms->push_back(copy);
  }
  delete subSplit;
  // now, split right term:
  subSplit = t2->splitPooling();
  for(vector<Term *>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
  {
    // add a copy of this to additionalAtoms with the right term replaced by the current alternative *subSplitIt:
    Term* backup = this->t2;
    this->t2 = NULL;
    BuiltinAtom* copy = this->clone();
    this->t2 = backup;
    copy->t2 = *subSplitIt;
    additionalAtoms->push_back(copy);
  }
  delete subSplit;
  
  vector<T*>* returnVector = new vector<T*>();
  for(vector<BuiltinAtom *>::iterator addit = additionalAtoms->begin(); addit!=additionalAtoms->end(); ++addit) // iterate over the additional atoms
  {
    returnVector->push_back(*addit);
    vector<T*>* nonpoolingTuples = (*addit)->splitPooling<T>(); // recursive pooling splitting
    returnVector->insert(returnVector->end(), nonpoolingTuples->begin(), nonpoolingTuples->end()); // adds all results to returnVector
    delete nonpoolingTuples;
  }
  delete additionalAtoms;
  updatePublicVariables();
  return returnVector;
}

vector<NafLiteral*>* BuiltinAtom::splitPoolingNaf()
{
  return splitPooling<NafLiteral>();
}

vector<BodyElement*>* BuiltinAtom::splitPooling()
{
  return splitPooling<BodyElement>();
}

vector<AggregateElement*>* AggregateElement::splitPooling()
{
  vector<AggregateElement*>* additionalElements = new vector<AggregateElement*>();

  // first, regard the term tuple:
  if(termTuple != NULL)
  {
    vector<TermTuple*>* subSplit = termTuple->splitPooling();
    for(vector<TermTuple *>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
    {
      // add a copy of this to additionalElements with the termTuple replaced by the current alternative *subSplitIt:
      TermTuple* backup = this->termTuple;
      this->termTuple = NULL;
      AggregateElement* copy = this->clone();
      this->termTuple = backup;
      copy->termTuple = *subSplitIt;
      additionalElements->push_back(copy);
    }
    delete subSplit;
  }
  
  // now, regard the dependency literals:
  vector<LiteralBody *> *subSplit = dependencyLiterals->splitPooling();
  for(vector<LiteralBody *>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
  {
    // add a copy of this to additionalElements with the dependencyLiterals replaced by the current alternative *subSplitIt:
    LiteralBody* backup = this->dependencyLiterals;
    this->dependencyLiterals = NULL;
    AggregateElement* copy = this->clone();
    this->dependencyLiterals = backup;
    copy->dependencyLiterals = *subSplitIt;
    additionalElements->push_back(copy);
  }
  delete subSplit;

  vector<AggregateElement*>* returnVector = new vector<AggregateElement*>();
  for(vector<AggregateElement *>::iterator addit = additionalElements->begin(); addit!=additionalElements->end(); ++addit) // iterate over the additional elements
  {
    returnVector->push_back(*addit);
    vector<AggregateElement*>* nonpoolingElements = (*addit)->splitPooling(); // recursive pooling splitting
    returnVector->insert(returnVector->end(), nonpoolingElements->begin(), nonpoolingElements->end()); // adds all results to returnVector
    delete nonpoolingElements;
  }
  delete additionalElements;
  return returnVector;
}

vector<BodyElement*>* Aggregate::splitPooling()
{
  // splits all poolings inside the aggregate by introducing new aggregate elements into this.
  // then, the guards are considered. If a guard has pooling, the returning vector will be non-empty.
  
  // first, split the elements *inside* this aggregate:
  vector<AggregateElement*>* splitElements = new vector<AggregateElement*>();
  for (vector<AggregateElement*>::iterator elementsIt = elements->begin(); elementsIt!=elements->end(); ++elementsIt)
  {
    vector<AggregateElement*>* sSplit = (*elementsIt)->splitPooling();
    splitElements->push_back(*elementsIt);
    splitElements->insert(splitElements->end(), sSplit->begin(), sSplit->end());
    delete sSplit;
  }
  delete this->elements;
  this->elements = splitElements;

  //now, handle guard poolings as usual, splitting into multiple aggregates:
  vector<Aggregate*>* additionalAggregates = new vector<Aggregate*>();
  
  // first, split left guard:
  if (guard_l != NULL)
  {
    vector<Guard*>* subSplit = guard_l->splitPooling();
    for(vector<Guard *>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
    {
      // add a copy of this to additionalAggregates with the left guard replaced by the current alternative *subSplitIt:
      Guard* backup = this->guard_l;
      this->guard_l = NULL;
      Aggregate* copy = this->clone();
      this->guard_l = backup;
      copy->guard_l = *subSplitIt;
      additionalAggregates->push_back(copy);
    }
    delete subSplit;
  }
  // now, split right guard:
  if (guard_r != NULL)
  {
    vector<Guard*>* subSplit = guard_r->splitPooling();
    for(vector<Guard *>::iterator subSplitIt = subSplit->begin(); subSplitIt!=subSplit->end(); ++subSplitIt)
    {
      // add a copy of this to additionalAggregates with the right guard replaced by the current alternative *subSplitIt:
      Guard* backup = this->guard_r;
      this->guard_r = NULL;
      Aggregate* copy = this->clone();
      this->guard_r = backup;
      copy->guard_r = *subSplitIt;
      additionalAggregates->push_back(copy);
    }
    delete subSplit;
  }
  
  vector<BodyElement*>* returnVector = new vector<BodyElement*>();
  for(vector<Aggregate *>::iterator addit = additionalAggregates->begin(); addit!=additionalAggregates->end(); ++addit) // iterate over the additional aggregates
  {
    returnVector->push_back(*addit);
    vector<BodyElement*>* nonpoolingTuples = (*addit)->splitPooling(); // recursive pooling splitting
    returnVector->insert(returnVector->end(), nonpoolingTuples->begin(), nonpoolingTuples->end()); // adds all results to returnVector
    delete nonpoolingTuples;
  }
  delete additionalAggregates;
  updatePublicVariables();
  return returnVector;
}

vector<LiteralBody*>* LiteralBody::splitPooling() // normal behaviour (like most other classes)
{
  vector<LiteralBody*>* additionalLiteralBodies = new vector<LiteralBody*>();
  
  // split literals:
  for(vector<NafLiteral *>::iterator literalsit = literals->begin(); literalsit!=literals->end(); ++literalsit) // iterate over the literals
  {
    vector<NafLiteral*>* res = (*literalsit)->splitPoolingNaf(); // split the current literal
    for(vector<NafLiteral *>::iterator altit = res->begin(); altit!=res->end(); ++altit) // iterate over the alternative literals of the current literal
    {
      // add a copy of this to additionalLiteralBodies with the current literal replaced by the current alternative:
      NafLiteral* backup = *literalsit;
      *literalsit = *altit;
      LiteralBody* copy = this->clone();
      delete *altit;
      *literalsit = backup;
      additionalLiteralBodies->insert(additionalLiteralBodies->begin(), copy);
    }
    delete res;
  }
  
  vector<LiteralBody*>* returnVector = new vector<LiteralBody*>();
  for(vector<LiteralBody *>::iterator addit = additionalLiteralBodies->begin(); addit!=additionalLiteralBodies->end(); ++addit) // iterate over the additional literal bodies
  {
    returnVector->push_back(*addit);
    vector<LiteralBody*>* nonpoolingBodies = (*addit)->splitPooling(); // recursive pooling splitting
    returnVector->insert(returnVector->end(), nonpoolingBodies->begin(), nonpoolingBodies->end()); // adds all results to returnVector
    delete nonpoolingBodies;
  }
  delete additionalLiteralBodies;
  calcVariables();
  return returnVector;
}

