/*
Copyright 2015-2016, Manuel Bichler
WWW: <http://dbai.tuwien.ac.at/proj/lpopt/>

This file is part of lpopt.

lpopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

lpopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lpopt.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "StatementOptimizationAlgorithm.hpp"
#include "../asp-domain/safetyCharacterization.hpp"
#include <algorithm>

using namespace std;
using namespace lpopt;

StatementOptimizationPartialSolution::StatementOptimizationPartialSolution() : createdStatements(), temporary(NULL), ignoreHeadTags()
{
}

StatementOptimizationPartialSolution::~StatementOptimizationPartialSolution()
{
	for (auto it = createdStatements.begin(); it != createdStatements.end(); ++it)
	{
		delete *it;
	}
	if(temporary) delete temporary;
}

StatementOptimizationAlgorithm::StatementOptimizationAlgorithm()
{
}

StatementOptimizationAlgorithm::~StatementOptimizationAlgorithm()
{
}

list<Statement*>* StatementOptimizationAlgorithm::optimize(const ITreeDecomposition &decomposition, HypergraphableStatement &instance) const
{
	unique_ptr<varset> originalStatementsVars(instance.stmt().variables());
	unique_ptr<StatementOptimizationPartialSolution> sol(evaluateNode(decomposition.root(), decomposition, instance, instance.stmt().clone(), *originalStatementsVars));
	list<Statement*>* toRet = new list<Statement*>();
	toRet->splice(toRet->end(), sol->createdStatements);
	return toRet;
}

StatementOptimizationPartialSolution* StatementOptimizationAlgorithm::evaluateNode(
				vertex_t node,
				const ITreeDecomposition &decomposition,
				HypergraphableStatement &instance,
				Statement* workingStatement,
				const varset& originalStatementsVars) const
{
	StatementOptimizationPartialSolution* toReturn;
	list<ClassicalLiteral*> childTemporaries;

	// gather children's partial solutions (bottom-up traversal of tree)
	ConstCollection<vertex_t> children = decomposition.children(node);
	if (children.size() == 0) toReturn = new StatementOptimizationPartialSolution();
	else
	{ // take first child's partialSolution and suck in all further
		toReturn = evaluateNode(*(children.begin()), decomposition, instance, workingStatement, originalStatementsVars);
		childTemporaries.push_back(toReturn->temporary);
		toReturn->temporary = NULL;
		for (auto it = ++(children.begin()); it != children.end(); ++it)
		{
			unique_ptr<StatementOptimizationPartialSolution> part(evaluateNode(*it, decomposition, instance, workingStatement, originalStatementsVars));
			toReturn->createdStatements.splice(toReturn->createdStatements.end(), part->createdStatements); // combine children's created statements in new partial solution
			childTemporaries.push_back(part->temporary);
			part->temporary = NULL;
			toReturn->ignoreHeadTags.insert(part->ignoreHeadTags.begin(), part->ignoreHeadTags.end());
		}
	}


	if (node != decomposition.root())
	{
		// which varaibles are included in this bag? ->thisVars
		varset thisVars;
		const vector<vertex_t>& bag = decomposition.bagContent(node);
		for (auto it = bag.begin(); it != bag.end(); ++it)
		{
			thisVars.insert(instance.getVariableById(*it));
		}

		// those body elements of the base rule that are in this node: (go to bodyPart)
		vector<BodyElement*>* bodyPart = workingStatement->extractBodyWithoutGlobvarUpdate(thisVars);
		Body* body = new Body(bodyPart);
		// add children's temporary literals:
		for (auto it = childTemporaries.begin(); it != childTemporaries.end(); ++it)
		{
			body->add(*it);
		}
		
		// create temporary predicate:
		auto intersectionVertices = decomposition.rememberedVertices(decomposition.parent(node), node);
		set<Variable*, var_cmp> tmpPredVars;
		for (auto it = intersectionVertices.begin(); it != intersectionVertices.end(); ++it)
		{
			tmpPredVars.insert(instance.getVariableById(*it)->clone());
		}

		// if we ignored head variables while decomposing, include head variables intersecting with this node's vertices in intersection and all those that have been tagged along:
		if (instance.ignoreHead())
		{
			for (auto it = toReturn->ignoreHeadTags.begin(); it != toReturn->ignoreHeadTags.end(); ++it)
			{
				Variable* newVar = (*it)->clone();
				if (!tmpPredVars.insert(newVar).second) // if the var was already contained
					delete newVar;
			}
			unique_ptr<varset> headVars(instance.stmt().headVars());
			vector<const Variable *> headBagIntersection;
			set_intersection(headVars->begin(), headVars->end(), thisVars.begin(), thisVars.end(), headBagIntersection.begin(), var_cmp());
			for (auto it = headBagIntersection.begin(); it != headBagIntersection.end(); ++it)
			{
				Variable* newVar = (*it)->clone();
				if (!tmpPredVars.insert(newVar).second) // if the var was already contained
					delete newVar;
				toReturn->ignoreHeadTags.insert(*it);
			}
		}

		ClassicalLiteral* tempPred;
		if (tmpPredVars.empty()) // no arguments in predicate / arity 0
			tempPred = new ClassicalLiteral(instance.gen().get(), false, false);
		else
		{
			vector<Term*>* tempPredTerms = new vector<Term*>(tmpPredVars.begin(), tmpPredVars.end());
			tempPred = new ClassicalLiteral(instance.gen().get(), false, false, new TermTuples(new vector<TermTuple *>(1, new TermTuple(tempPredTerms))));
		}

		// deal with unsafe variables:
		SafetyCharacterization* safety = NULL;
		while(!(safety = body->safety(originalStatementsVars))->unsafeVars().empty())
		{
			// for each unsafe var, create new Rule and add the safety predicate to the body: TODO: wouldn't it be more efficient to generate only one safety predicate for all unsafe variables?
			// choose those variable first that cannot be made safe with the current body and recheck safety afterwards.
			unique_ptr<varset> cands(safety->nonSafeable());
			const Variable* cand;
			if (!cands->empty()) cand = *(cands->begin());
			else cand = *(safety->unsafeVars().begin());
			unique_ptr<vector<const BodyElement*>> safemakers(instance.stmt().selectSafe(*cand));
			ClassicalLiteral* safetyPred = new ClassicalLiteral(&(instance.gen().get()->append("_safe")), false, false, new TermTuples(new vector<TermTuple *>(1, new TermTuple(new vector<Term*>(1, new Variable(*cand))))));
			body->add(safetyPred);
			vector<BodyElement*>* bodyElements = new vector<BodyElement*>();
			for (vector<const BodyElement*>::iterator bit = safemakers->begin(); bit != safemakers->end(); ++bit)
				bodyElements->push_back((*bit)->clone());
			toReturn->createdStatements.push_back(new Rule(new Disjunction(new vector<ClassicalLiteral*>(1, safetyPred->clone())), new Body(bodyElements))); // TODO: new safety rule may be split up as well!!!
			delete safety;
		}
		delete safety;

		// make new rule:
		toReturn->createdStatements.push_back(new Rule(new Disjunction(new vector<ClassicalLiteral *>(1, tempPred->clone())), body));
    
		toReturn->temporary = tempPred;
	}

	else // node == decomposition.root()
	{
		// since we used extractBodyWithoutGlobvarUpdate:
		workingStatement->calcVariables();
		// now, workingStatement consists only of variable-free body elements and the original head.
		// since we are in head node, there are no unsafes in our body.
		// add the temps to the body, then we are finished:
		for (auto it = childTemporaries.begin(); it != childTemporaries.end(); ++it)
		{
			workingStatement->addToBody(*it);
		}
		toReturn->createdStatements.push_back(workingStatement);
	}
	return toReturn;
}


